#ifndef _CWS_
#define _CWS_
//#include <iostream>
//#include <string>

//using namespace std;

//void getCWS(string fileName, double *rcb, int sketchSize, int dimension);
void getCWS(double *r, double *c, double *b, int sketchSize, int dimension);


#endif
