#ifndef _DISTANCE_
#define _DISTANCE_

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <stdint.h>
#include <cmath>

using namespace std;

double getWJD(uint32_t *setA, uint32_t *setB, double *weightA, double *weightB, int sketchSize);















#endif
